{{ SumoIO.spin }}

''  Provides low level SPI drivers for interfacing with the analog and digital I/O
''  on the robot controller
''

CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 6_000_000

VAR
  long  cog
    
PUB start(buff_ptr) : okay

'' Start driver - starts a cog
'' returns false if no cog available
'' may be called again to change settings
''
                  
  stop
  return cog := cognew(@entry, buff_ptr) + 1


PUB stop

'' Stop driver - frees a cog

  if cog
    cogstop(cog~ - 1)

  
DAT

'*********************************************************
'* Assembly language SPI driver for MCP3208 and MCP23S08 *
'*********************************************************

                        org
'
'
' Read initialization parameters
'
entry

' Setup I/O pin directions
                        
                        or      outa, clk_pin
                        or      outa, cs_adc_pin
                        or      outa, cs_xio_pin
                        
                        or      dira, cs_adc_pin
                        or      dira, cs_xio_pin
                        or      dira, clk_pin
                        or      dira, dout_pin

                        andn    dira, din_pin
                        andn    dira, int_pin


' Setup hardware counters to enforce maximum refresh rates
                        mov     ctra, ctr_mode
                        mov     ctrb, ctr_mode
                        mov     frqa, #1
                        mov     frqb, #1
                        mov     phsa, #0
                        mov     phsb, #0


' Setup MCP23S08
                        mov     cmd, #$01      ' Setup I/O Polarity for Normal
                        mov     bits, #$00
                        call    #write_xio

                        mov     cmd, #$02      ' Disable Interupt on Change
                        mov     bits, #$00      
                        call    #write_xio
                        
                        mov     cmd, #$06      ' Set Inputs to have internal pull-up
                        mov     bits, #$FF      
                        call    #write_xio
                        
                        
' Perform conversions at the specified maximium refresh rates

' if ADC counter has expired then read values from A2D converter
' for the enabled channel.

' if XIO counter has expired then read pin outputs and direction
' from cog RAM and update MCP23S08 if changes have occured.
' Check the status of the interrupt pin and read new input values
' if any change has occured.
'
                        
main_loop
                        mov     p0, par
                        rdlong  adc_rate, p0
                        add     p0, #4
                        rdlong  xio_rate, p0
                        
                        mov     t1, phsa
                        cmp     adc_rate, t1 wc
              if_c      call    #adc_loop

                        mov     t1, phsb
                        cmp     xio_rate, t1 wc
              if_c      call    #xio_loop

                         
                        jmp     #main_loop

' Read Analog Inputs
adc_loop
                        mov     phsa, #0
                        
                        mov     p0,   par               ' Read Analog I/O Settings
                        add     p0,   #12
                        rdbyte  analog_mode, p0
                        add     p0,   #1
                        rdbyte  analog_en,   p0
                        add     p0,   #1
                        rdword  analog_rate, p0

                        add     average_cnt, #1
                        cmp     average_cnt, analog_rate wc, wz
              if_ae     mov     average_cnt, #0

                        mov     p0,   par
                        add     p0,   #16
                        mov     p1,   p0
                        add     p1,   #16

                                           
                        mov     t0,   #0
                        mov     t4,   #8
:loop
                        test    analog_en, #1 wz
                 if_z   jmp     #:next
                        call    #read_adc

                        test    analog_mode, #1 wz      ' Check if value is to be averaged
                 if_z   jmp     #:save

                        rdlong  t2,   p1                ' Read old sum
                        add     t2,   t1
                        cmp     average_cnt, #0 wz
                 if_nz  wrlong  t2,   p1
                 if_nz  jmp     #:next                  ' Update sum nothing else to save

                        mov     x,    t2                ' Calculate Average
                        mov     y,    analog_rate
                        call    #divide

                        mov     t1,   x
                        and     t1,   mask12
                        mov     t2,   #0
                        wrlong  t2,   p1

:save                   wrword  t1,   p0               ' Write immediate or averaged value
                 
:next
                        add     p0,   #2
                        add     p1,   #4

                        shr     analog_en, #1
                        shr     analog_mode, #1

                        add     t0,   #1
                        djnz    t4,   #:loop

                        mov     p0,   par
                        add     p0,   #8
                                                     
                        rdlong  t1,   p0
                        cmp     average_cnt, #0 wz

                if_nz   add     t1,   #2                
                if_z    add     t1,   #1
                
                        wrlong  t1,   p0

adc_loop_ret            ret

xio_config
                        mov     digital_mode, t0
xio_config_ret          ret


xio_loop
                        mov     phsb, #0
                        mov     p0,   par
                        add     p0,   #64
                        mov     p1,   p0
                        add     p1,   #4

                        rdlong  t0,   p1
                        cmp     t0,   digital_mode wz
                if_nz   call    #xio_config

                        add     p1,   #5
                        
                        mov     cmd, #$00               'Set I/O Pin Direction
                        rdbyte  bits, p1
                        add     p1,   #1
                        call    #write_xio

                        mov     cmd,  #$09              'Read Input Register
                        call    #read_xio

                        wrbyte  bits, p1
                        add     p1,   #1

                        mov     cmd, #$0A               'Set Output Register
                        rdlong  bits, p1
                        call    #write_xio

                        rdlong  t0,   p0                'Update Counter
                        add     t0,   #1
                        wrlong  t0,   p0
xio_loop_ret            ret

reverse_4lsb
                        mov     t2,   t1                
                        rev     t2,   #4
                        shr     t2,   #24
                        
                        andn    t1,   #%1111
                        or      t1,   t2
reverse_4lsb_ret        ret


write_xio
                        or      outa, clk_pin
                        or      outa, cs_adc_pin
                        andn    outa, cs_xio_pin

                        mov     c1, cnt
                        add     c1, t_clk
                        waitcnt c1, t_clk

                        mov     t2, #%0100_0000  
                        call    #write
                        mov     t2, cmd
                        call    #write
                        mov     t2, bits
                        call    #write

                        or      outa, cs_xio_pin
write_xio_ret           ret

read_xio
                        or      outa, clk_pin
                        or      outa, cs_adc_pin
                        andn    outa, cs_xio_pin

                        mov     c1, cnt
                        add     c1, t_clk
                        waitcnt c1, t_clk

                        mov     t2, #%0100_0001  
                        call    #write
                        mov     t2, cmd
                        call    #write

                        call    #read
                        mov     bits, t2

                        or      outa, cs_xio_pin
read_xio_ret            ret

read
                        mov     t1, #8
                        mov     t2, #0

:loop                   call    #strobe_clk
                        
                        shl     t2, #1
                        mov     t4, ina
                        test    t4, din_pin wz
                        muxnz   t2, #1

                        djnz    t1, #:loop
                        
read_ret                ret

write
                        mov     t1, #8
:loop                   test    t2, #$80 wz             ' Shift data high bit first
                        muxnz   outa, dout_pin
                        shl     t2, #1

                        call    #strobe_clk
                                                              
                        djnz    t1, #:loop
write_ret               ret



'Read from MCP3208 ADC
'   t0 - channel to read
'   t1 - 12 bit result

read_adc

                        
                        or      outa, clk_pin
                        or      outa, cs_xio_pin
                        andn    outa, cs_adc_pin

                        mov     c1, cnt
                        add     c1, t_clk
                        waitcnt c1, t_clk
                        

                        or      outa, dout_pin          ' Start Bit
                        call    #strobe_clk
              
                        or      outa, dout_pin          ' 1 - Single/ 0 - Diff Mode
                        call    #strobe_clk

                        test    t0,   #%100  wz         ' 3 bit channel selection
                        muxnz   outa, dout_pin                                           
                        call    #strobe_clk

                        test    t0,   #%010  wz
                        muxnz   outa, dout_pin                                           
                        call    #strobe_clk
                                                                                       
                        test    t0,   #%001  wz
                        muxnz   outa, dout_pin                                           
                        call    #strobe_clk

                        call    #strobe_clk             ' Sample Strobe

                        mov     t1,   #0
                        mov     t2,   #13
                                                
:loop                   andn    outa, clk_pin           
                        waitcnt c1, t_clk
                        or      outa, clk_pin

                        shl     t1, #1
                        mov     t3, ina
                        test    t3, din_pin wz
                        muxnz   t1, #1

                        waitcnt c1, t_clk
                        djnz    t2, #:loop

                        or      outa, cs_adc_pin
read_adc_ret            ret

       
strobe_clk
                        mov     c1, cnt
                        add     c1, t_clk
                        andn    outa, clk_pin           
                        waitcnt c1, t_clk
                        or      outa, clk_pin
                        waitcnt c1, t_clk
strobe_clk_ret          ret
                                                        
                        
'
' Get a parameter
'
param                   rdlong  t3,t1                   'get parameter into t3
                        add     t1,#4                   'point to next parameter
param_ret               ret

{ Divide x[31..0] by y[15..0] (y[16] must be 0)
  on exit, quotient is in x[15..0] and remainder is in
  x[31..16] }

divide        shl       y,          #15
              mov       t,          #16
:loop         cmpsub    x,          y wc
              rcl       x,          #1
              djnz      t,          #:loop
divide_ret    ret

'
'
' Initialized data
'
dlsb                    long    1 << 9
mask12                  long    $FFF

cs_adc_pin              long    1 << 19
cs_xio_pin              long    1 << 21

'led0                   long    1 << 31
'led1                   long    1 << 30

clk_pin                 long    1 << 23
din_pin                 long    1 << 20
dout_pin                long    1 << 22
int_pin                 long    1 << 24

ctr_mode                long    %0_11111_000_00000000_000000_000_000000

t_clk                   long    48

digital_mode            long    %00000000_00000000_00000000_00000000
 
'
'
' Uninitialized data
'
t0                      res     1
t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1

c1                      res     1

p0                      res     1
p1                      res     1

x                       res     1
y                       res     1
t                       res     1

command                 res     1
stream                  res     1

adc_bits                res     1
adc_rate                res     1

xio_bits                res     1 
xio_rate                res     1                       

last_output             res     1
last_dir                res     1

cmd                     res     1
bits                    res     1

analog_mode             res     1
analog_en               res     1
analog_rate             res     1

average_cnt             res     1   